/*------------------------------------------------------*/
/* A biomorph generator inspired by Richard Dawkins, 	*/
/* With nervous nets inspired by Mark Tilden		*/
/*------------------------------------------------------*/

/* Initial width and height */

#define WWIDTH 750
#define WHEIGHT 750

#define sign(a)		((a) < 0 ? -1 : 1)
#define abs(a)		((a) < 0 ? -(a) : (a))

/* 3-D point position in integer units */

typedef struct {
   int x;
   int y;
   int z;
} Point3D;

/* Types of conditional operators */

#define EQUAL		0
#define NOT_EQUAL	1
#define LESS_THAN	2
#define MORE_THAN	3

/*------------------------------------------------------*/
/* Type record 						*/
/* The type itself (type name, a string) is the key to	*/
/* the type hashtable entry.				*/
/*------------------------------------------------------*/

typedef struct _typerecord {
   float r, g, b, a;		/* RGBA color to use to render the type */
   unsigned int idx;		/* Sequential index for processing optimization */
} Type;

/*------------------------------*/
/* DNA table entries 		*/
/*------------------------------*/


/* This is saved in a hash table according to cell type */
/* (hash tag is cell typename (string))			*/

typedef struct _dnacond *DNACondptr;
typedef struct _dnacond {
   Type *type;		/* Type being evaluated */
   char operator;	/* Operator (see above) */
   char number;		/* Number of cells (1 to 12) */
   DNACondptr next;
} DNACond;

typedef struct _dnarule *DNARuleptr;
typedef struct _dnarule {
   DNACond *condition;	/* Linked list of conditions */
   Type *result;	/* Pointer to Type record to change to on next cycle */
   DNARuleptr next;
} DNARule;

typedef struct _dnaruleset {
   Tcl_HashTable DNARuleHash;
} DNARuleset;

extern Tcl_HashTable DNARuleSets;

/*------------------------------------------------------*/
/* Type record hash					*/
/* Hash tag is generated from type value		*/
/*------------------------------------------------------*/

extern Tcl_HashTable TypeHashTable;

/*------------------------------------------------------*/
/* Cell record hash 					*/
/* Hash tag is generated from position coordinates	*/
/* (tag is a Point3D structure)				*/
/*------------------------------------------------------*/

extern Tcl_HashTable CellHashTable;

/*------------------------------------------------------*/
/* Cell record 						*/
/*------------------------------------------------------*/

typedef struct _cellrecord {
   Type *type;		/* pointer to cell type */
   DNARuleset *ruleset; /* pointer to set of DNA rules used by cell */
} Cell;

extern Type *SpaceType;

/* Event types as marked in the input data file */

#define CHECK   '?'
#define CHANGE	'!'

/*------------------------------*/
/* Event records		*/
/*------------------------------*/

typedef struct _changerecord *ChangeRecordptr;

typedef struct _changerecord {
   char phash[19];		// location of event (hashed)
   Type *type;			// type to be changed to
   DNARuleset *ruleset;		// ruleset to be changed to
   ChangeRecordptr next;
} ChangeRecord;

typedef struct _checkrecord *CheckRecordptr;

typedef struct _checkrecord {
   DNARuleset *ruleset;		// ruleset to use to check
   CheckRecordptr next;		// chained checks where conflicts occur
} CheckRecord;

/* The event stacks */

extern ChangeRecord *ChangeEventStack;
extern Tcl_HashTable CheckEventTable;

/* Other global variables used by multiple source files */

extern int verbose;
extern char zslice;
extern int zplane;
extern unsigned int num_types;
extern float viewangle[4];
extern float viewzoom;

/*--------------------------------------------------------------*/
/* Forward declarations						*/
/*--------------------------------------------------------------*/

extern char *PointToHash(Point3D *point);
extern Point3D *HashToPoint(char *phash);
